import 'package:global_state/src/flutter.dart' show protected;

import 'error.dart';

/// Interface required by [GSController] to notify changes in observeds.
///
/// Implemented by [GSElement].
abstract class Listener implements FastHashCode {
  @protected
  Set<Notifier> notifiers;

  /// Handles a change notified by an [Notifier].
  ///
  /// `postponeEventHandling` advices the listener to postpone the event
  /// handling.
  @protected
  onObservedChange(Notifier notifier, [bool postponeEventHandling = false]);
}

/// Interface necessary by [GSController] to register and notify
/// listeners of changes.
///
/// Implemented by [GSNotifier], [GSValue] and [GSMap].
abstract class Notifier implements FastHashCode {
  /// Set used by the controller to store the listeners of the notifier.
  @protected
  Set<Listener> listeners;

  /// Notifies the controller about a read event.
  notifyRead();

  /// Notifies the listeners about a change event.
  notifyChange();
}

/// Class that subscribes [Listener] to [Notifier] instances.
class GSController {
  static Listener _activeListener;

  /// The object that is currently listening to [Notifier] instances.
  static Listener get activeListener => _activeListener;

  static bool get isListening => _activeListener != null;

  static startListening(Listener listener) {
    if (isListening) {
      stopListening();
    }
    _activeListener = listener;
  }

  static void stopListening() {
    assert(_activeListener != null);
    _commitSubscriptions();
    assert(_currentNotifiers.isEmpty);
    _activeListener = null;
  }

  static Set<Notifier> _currentNotifiers = Set();

  static registerCurrentNotifier(Notifier notifier) {
    assert(isListening);
    notifier.listeners ??= Set();
    _currentNotifiers.add(notifier);
  }

  static _commitSubscriptions() {
    final listener = _activeListener;
    final previousNotifiers = _activeListener.notifiers;
    assert(() {
      if (_currentNotifiers.isNotEmpty &&
          (previousNotifiers == null || previousNotifiers.isEmpty)) {
        GSController._debugSubscribedListeners.add(_activeListener);
      }
      return true;
    }());
    if (previousNotifiers == null) {
      if (_currentNotifiers.isNotEmpty) {
        for (var notifier in _currentNotifiers) {
          notifier.listeners.add(listener);
        }
        listener.notifiers = _currentNotifiers;
        _currentNotifiers = Set();
      }
      return;
    }
    // Most element rebuilds should read the same keys, a quick check is done
    // to handle that case.
    else if (previousNotifiers.length == _currentNotifiers.length &&
        previousNotifiers.containsAll(_currentNotifiers)) {
      _currentNotifiers.clear();
      return;
    }

    // Registers listener in new notifiers.
    for (var newNotifier in _currentNotifiers) {
      if (!previousNotifiers.remove(newNotifier)) {
        newNotifier.listeners.add(listener);
      }
    }
    // Unregister listener from old notifiers (previousNotifiers - currentNotifiers).
    for (var oldNotifier in previousNotifiers) {
      oldNotifier.listeners.remove(listener);
    }
    listener.notifiers = _currentNotifiers;
    _currentNotifiers = previousNotifiers..clear();
  }

  /// Flag that prevents notification events from triggering an assertion error
  /// when listening.
  static bool debugAllowNotificationsWhenListening = false;

  static notifyChangeToListeners(Notifier notifier,
      [bool advicePostponing = false]) {
    assert(() {
      if (isListening &&
          !advicePostponing &&
          !debugAllowNotificationsWhenListening) {
        gsError('`$activeListener` is listening (a widget is building) '
            'while setting value of the [Notifier] $notifier. '
            '[Notifier] instances like [GSMap] cannot be modified '
            'within a build method.\n'
            'In GS widgets, initContext can be used to initialize values '
            'values of GS maps, lists or values.');
        return false;
      }
      return true;
    }());
    for (var listener in notifier.listeners) {
      listener.onObservedChange(notifier, advicePostponing);
    }
  }

  static unsubscribeListener(Listener listener) {
    assert(!isListening);
    for (var notifier in listener.notifiers) {
      assert(notifier.listeners.contains(listener));
      notifier.listeners.remove(listener);
    }
    listener.notifiers.clear();
    assert(() {
      GSController._debugSubscribedListeners.remove(listener);
      return true;
    }());
  }

  static unsubscribeNotifier(Notifier notifier) {
    for (var listener in notifier.listeners) {
      listener.notifiers.remove(notifier);
    }
    notifier.listeners.clear();
  }

  static final _debugSubscribedListeners = Set<Listener>();

  static int get debugSubscribedListenersCount =>
      _debugSubscribedListeners.length;

  static debugReset() {
    for (var listener in _debugSubscribedListeners.toList()) {
      unsubscribeListener(listener);
    }
  }

  static bool debugContainsListener(Listener listener) =>
      _debugSubscribedListeners.contains(listener);
}

/// Class to retrieve a fast hashCode.
///
/// The inherited Dart implementation of hashCode is very slow, it takes 10
/// times longer than using a fixed int. This causes a big performance drop in
/// hash data structures like Set or Map. Implementations os [Notifier]
/// and [Listener] should include this mixin.
mixin FastHashCode {
  static int _lastId = 0;

  /// The value returned by [hashCode].
  ///
  /// Implementers of [FastHashCode] interface that do not include it as mixin
  /// should override [hashcode] to return this value.
  final hashId = _lastId++;

  int get hashCode => hashId;
}

enum ObservedStatus {
  active,
  defunct,
}

/// Mixin that connects observeds with [Listener] instances that use
/// [GSController].
///
/// Mixed in by [GSNotifier], base class of [GSValue] and [GSMap].
abstract class NotifierMixin implements Notifier {
  var _debugStatus = ObservedStatus.active;

  Listener get activeListener => GSController._activeListener;
  bool get controllerIsListening => GSController.isListening;

  @protected
  Set<Listener> listeners;

  bool get hasRegisteredListeners => listeners != null && listeners.isNotEmpty;

  void notifyRead() {
    assert(_debugStatus != ObservedStatus.defunct);
    if (GSController.isListening) {
      GSController.registerCurrentNotifier(this);
    }
  }

  void notifyChange([bool postponeNotificationHandling = false]) {
    assert(_debugStatus != ObservedStatus.defunct);
    if (listeners != null) {
      if (listeners.isNotEmpty) {
        GSController.notifyChangeToListeners(
            this, postponeNotificationHandling);
      } else if (postponeNotificationHandling) {
        GSController.notifyChangeToListeners(this, true);
      }
    }
  }

  void forgetListeners() {
    assert(_debugStatus != ObservedStatus.defunct);
    if (hasRegisteredListeners) {
      GSController.unsubscribeNotifier(this);
    }
  }

  /// Dispose can be invoked when this notifier is not going to be used again.
  void dispose() {
    assert(_debugStatus == ObservedStatus.active);
    forgetListeners();
    assert(() {
      _debugStatus = ObservedStatus.defunct;
      return true;
    }());
  }
}
